* OutputClass.prg	(c) Peisch Custom Software, 2006

*   You may include or modify any of the code included in this demo as you see fit.
*   You may not claim it as your own, nor try to enforce a copyright.
*   All samples in this demo are provided "as is" with no implied guarantees nor waranties.

DEFINE CLASS Output as Custom
	*   A generic and centralized place for calling reports and output

	* You'll probably want these properties regardless.
	* They're properties you're likely to want to change or read.
	cReportMode = "NATIVE"		&& Which tool to use for output
	cProgressMess = ''			&& Optional display message
	lCancelProcessing = .F.		&& Flag if user canceled confirmation of output
	cDefaultOutputFile = ''		&& When sending output to a file, where should it go?
	lDontConfirmOutput = .F.	&& If .T. send output to cDefaultOutputFile without confirmation
	cOutput = ''				&& Type of output, e.g. PDF, DOC, PREVIEW, etc.
	lDebugMode = .F.			&& Flag for turning on debugging

	* These are properties required for some third party tools
	oRpt = .Null.				&& Object reference to 3rd party tool class
	lXFRXBookmarks = .F.		&& For XFRX output only
	oScripts = .Null.			&& Used internally for scripts in XFRX only
	
	* Necessary, but only used internally
	cPrinter = ''				&& For internal use only
	cOldPrinter = ''			&& For internal use only
	oListeners = .Null.			&& Collection of listeners
	
	DIMENSION aRpts2Run[1]		&& Place for queuing reports

	*-----------------------------------------------------
	FUNCTION Init
	
	This.oListeners = CREATEOBJECT("collection")

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION InitRptMode(tcOutput as String)
	* Called when report mode is switched on the form with the option buttons
	* tcOutput = determines where to send the report, e.g. Preview, PDF, HTML

	IF EMPTY(tcOutput)
		This.cOutput = 'PREVIEW'
	ELSE
		This.cOutput = UPPER(tcOutput)
	ENDIF 

	DO CASE 
	CASE This.cReportMode = "XFRX"
		this.InitXFRX()
	CASE this.cReportMode = "NATIVE"
		this.InitNative()
	CASE this.cReportMode = "WORD TABLE"
		This.InitCur2Word()
	OTHERWISE
		MESSAGEBOX("I don't know how to handle this report mode!")
	ENDCASE 
	
	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION Destroy
	LOCAL lcClass

	IF This.cReportMode = "XFRX" AND NOT ISNULL(This.oScripts)
		* It's not necessary to unregister scripts, if we're going
		* to exit, but I'm just showing how to do it.
*!*			This.oRpt.UnRegisterAllScripts()
*!*			this.oScripts = .Null.
	ENDIF
	this.oRpt = .Null.
	
	IF NOT EMPTY(this.cOldPrinter)
		SET PRINTER TO (this.cOldPrinter)
	ENDIF
	
	ENDFUNC
	*-----------------------------------------------------
	FUNCTION QueueRpt(tcReportName as String, toListeners as Object)
	* tcReportName = the name of the FRX file
	* taListeners  = Optional parameter for an array of listeners

	IF NOT EMPTY(This.cProgressMess)
		WAIT WINDOW NOWAIT This.cProgressMess
	ENDIF

	This.StoreRpt(tcReportName, toListeners)

	WAIT CLEAR 

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION StoreRpt(tcReportName as String, toListeners as Object)

	IF ALEN(this.aRpts2Run) = 1
		lnRows = 1
	ELSE
		lnRows = ALEN(this.aRpts2Run,1)+1
	ENDIF 
	DIMENSION this.aRpts2Run[lnRows,3]
	this.aRpts2Run[lnRows,1] = tcReportName
	this.aRpts2Run[lnRows,2] = ALIAS()
	this.aRpts2Run[lnRows,3] = toListeners
	
	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION InitNative
	
	DO CASE 
*!*		CASE  This.cOutput = "PRINT"
*!*			this.cOldPrinter = SET("Print",3)
*!*			this.cPrinter = GETPRINTER()
*!*			IF EMPTY(this.cPrinter)
*!*				This.lCancelProcessing = .T.
*!*				RETURN .F.
*!*			ELSE
*!*				SET PRINTER TO (this.cPrinter)
*!*				This.lCancelProcessing = .F.
*!*			ENDIF 
	CASE This.cOutput = "HTML"
		IF This.lDontConfirmOutput
			lcOutputFile = This.cDefaultOutputFile
		ELSE 
			lcOutputFile = PUTFILE("HTML file name:",This.cDefaultOutputFile,"HTM")
			IF EMPTY(lcOutputFile)
				This.lCancelProcessing = .T.
				RETURN .F.
			ELSE
				This.cDefaultOutputFile = lcOutputFile
				This.lCancelProcessing = .F.
			ENDIF 
		ENDIF 
	ENDCASE

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION InitCur2Word

	IF This.lDontConfirmOutput
		lcOutputFile = This.cDefaultOutputFile
	else
		lcOutputFile = PUTFILE("Word file name:",This.cDefaultOutputFile,"DOC")
		IF EMPTY(lcOutputFile)
			This.lCancelProcessing = .T.
			RETURN .F.
		ELSE
			This.cDefaultOutputFile = lcOutputFile
			This.lCancelProcessing = .F.
		ENDIF 
	endif
	
	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION InitXFRX
	LOCAL lcOutputFile, lnRetVal

	DIMENSION This.aRpts2Run[1]
	This.aRpts2Run[1] = ''

	this.oRpt = xfrx("XFRX#LISTENER")

	DO CASE 
	CASE this.cOutput = 'PREVIEW'
		lcOutputFile = ''
	CASE this.lDontConfirmOutput
		lcOutputFile = This.cDefaultOutputFile
	OTHERWISE  
		lcOutputFile = PUTFILE(This.cOutput+" file name:",This.cDefaultOutputFile,This.cOutput)
	ENDCASE

	IF NOT EMPTY(lcOutputFile) OR this.cOutput = "PREVIEW"
		lnRetVal = this.oRpt.SetParams(lcOutputFile,,.F.,,.T.,,IIF(This.cOutput="PREVIEW",'XFF',This.cOutput))
*!*			This.RegisterXFRXScripts()
	
		IF lnRetVal = 0
			* Everything was Ok
			this.oRpt.ShrinkHeight(0)	&& Only affects HTML output

			IF This.cReportMode = "XFRX"
				this.oRpt.SetOtherParams("PRINT_BOOKMARKS",This.lXFRXBookmarks,1)	&& If HTML, the 1 parameter prints report bookmarks only
			ENDIF 
		ELSE
			MESSAGEBOX(This.oRpt.ErrorMessage(lnRetVal),16,"Report error")
		ENDIF 
	ELSE
		This.lCancelProcessing = .T.
	ENDIF

	RETURN lnRetVal

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION SendRptToOutput
	
	IF ALEN(This.aRpts2Run) = 1 OR TYPE('This.aRpts2Run[1,1]') <> "C"
		MESSAGEBOX("There is no data for your report selection.",64,"No data")
		RETURN 
	ENDIF 
	
	IF NOT EMPTY(This.cProgressMess)
		WAIT WINDOW NOWAIT This.cProgressMess
	ENDIF
	
	DO CASE 
	CASE This.cReportMode = "WORD TABLE"
		This.Cursor2Word()
	CASE this.cReportMode = "XFRX"
		this.OutputLoop()
		IF This.cOutput = "PREVIEW"
			* The preview form is modal so we must call wait clear before calling the form
			WAIT clear
			DO FORM Preview WITH this
		ELSE 
			WAIT clear
		ENDIF 
	CASE This.cReportMode = "NATIVE"
		IF This.cOutput = "PREVIEW"
			This.OutputLoop()
		ELSE
			* It's HTML
			This.NativeHTML()
		ENDIF 
	ENDCASE 

	* Clear out the array
	DIMENSION This.aRpts2Run[1]
	This.aRpts2Run[1] = ''
	
	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION InstantiateListeners(tnRptNum as Integer, toListener as Object)
	LOCAL lnListenerCount, lnStartCount, lnOffset

	* toListener is an optional parameter.  If it is passed it's assumed this is a listener
	* that should be added as the first in the chain, before the listeners in This.oListeners.

	* NOTE: We are assumming that all listener classes are based on the _ReportListener class
	* in _reportlistener.vcx (in the VFP home\ffc directory).  This is where the IsSuccessor
	* and Successor properties come from.

	* First clear out the oListeners property
	This.oListeners = .NULL.
	This.oListeners = CREATEOBJECT("Collection")

	IF TYPE('toListener') = "O"
		lnStartCount = 0
		lnOffset = 1
	ELSE
		lnStartCount = 1
		lnOffset = 0
	ENDIF 
	
	* Now use this property to store pointers to each of the listeners for the current report
	lnListenerCount = This.aRpts2Run[tnRptNum,3].Count
	FOR i = lnStartCount TO lnListenerCount
		IF i < 1
			This.oListeners.Add(toListener)
		ELSE 
			This.oListeners.Add(CREATEOBJECT(This.aRpts2Run[tnRptNum,3].Item[i]))
		ENDIF 

		* Set the successor info
		DO CASE 
		CASE i = 0
			* This is the object passed in toListner. Don't need to do anything.
			* We'll set the successor reference the next time through the loop.
		CASE i >= 1
			IF i = 1 AND TYPE('toListener') <> "O"
				* This is the first in the chain and there's no listener passed.
				* We'll set the successor reference on the next time through the loop.
			ELSE 
				This.oListeners.Item[i+lnOffset].IsSuccessor = .T.
				This.oListeners.Item[i-lnStartCount].Successor = This.oListeners.Item[i+lnOffset]
			ENDIF 
		ENDCASE 
	ENDFOR 
	
	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION Cursor2Word
	LOCAL loWordTables, loWord
	
	loWord = CREATEOBJECT("Word.Application")
	loWordTables = NEWOBJECT("WordTables","WordTables.prg")
	loWordTables.cWordFile = This.cDefaultOutputFile
	loWordTables.lShowStatus = .T.
	FOR i = 1 TO ALEN(This.aRpts2Run,1)
		loWordTables.nTable = i
		* Transform each alias in This.aRpts2Run into an array property of the WordTables class
		SELECT (This.aRpts2Run[i,2])
		DIMENSION loWordTables.aTableData[RECCOUNT(),FCOUNT()]
		loWordTables.aTableData = ''		&& Clear out the array each time through the loop
		SCAN
			FOR j = 1 TO FCOUNT()
				loWordTables.aTableData[RECNO(),j] = EVALUATE(FIELD(j))
			ENDFOR 
		ENDSCAN 
		loWordTables.WriteTable()
		IF EMPTY(loWordTables.cErrorMessage)
			* Success!
		ELSE 
			* There was an error.
			MESSAGEBOX(loWordTables.cErrorMessage)
		ENDIF 
	ENDFOR 
	loWord.Quit()
	
	* Open the file so we can see it
	IF FILE(loWordTables.cWordFile)
		DECLARE INTEGER ShellExecute IN SHELL32.DLL ;
		        INTEGER lnHWnd, ;
		        STRING lcAction, ;   
		        STRING lcFileName, ;
		        STRING lcExeParams, ;
		        STRING lcDefDir, ;
		        INTEGER lnShowWindow
		ShellExecute( 0, 'open', loWordTables.cWordFile, "", "", 1 )
	ENDIF 
	
	RELEASE loWordTables

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION NativeHTML
	
	This.oRpt = CREATEOBJECT("HtmlListener")
	This.oRpt.targetFileName = This.cDefaultOutputFile

	This.OutputLoop()
	
	* Open the file so we can see it
	IF FILE(This.cDefaultOutputFile)
		DECLARE INTEGER ShellExecute IN SHELL32.DLL ;
		        INTEGER lnHWnd, ;
		        STRING lcAction, ;   
		        STRING lcFileName, ;
		        STRING lcExeParams, ;
		        STRING lcDefDir, ;
		        INTEGER lnShowWindow
		ShellExecute( 0, 'open', This.cDefaultOutputFile, "", "", 1 )
	ENDIF 

	ENDFUNC
	*-----------------------------------------------------
	FUNCTION OutputLoop
	
	FOR lnRpt = 1 TO ALEN(This.aRpts2Run,1)
		* Instantiate any listeners required.
		* Zero out the count of This.oListeners if there are none.
		* Native HTML output requires the HTML listener, which we've stored in This.oRpt.
		* Send This.oRpt as the second parameter to add it to the beginning of the chain.
		This.InstantiateListeners(lnRpt,IIF(TYPE('This.oRpt') = "O",This.oRpt,.F.))

		SELECT (This.aRpts2Run[lnRpt,2])
		
		* Note: NORESET affects the page number that the CURRENT report starts with
		DO CASE 
		CASE lnRpt = 1
			* We want to make sure the page number is reset for the first report, but
			* only keep it from ejecting if there are more reports to run.
			IF ALEN(This.aRpts2Run,1) = 1
				lcClauses = ""
			ELSE
				lcClauses = "NOPAGEEJECT"
			endif
		CASE lnRpt = ALEN(This.aRpts2Run,1)
			lcClauses = "NORESET"
		OTHERWISE 
			lcClauses = "NOPAGEEJECT NORESET"
		ENDCASE	

		This.CallReport(lcClauses)
	ENDFOR 	

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION CallReport(tcClauses as String)
	LOCAL lcOldOnError

	IF EMPTY(tcClauses)
		tcClauses = ''
	ENDIF 

	DO CASE 
	CASE This.lDebugMode
		* We need to turn off any error handling for this to work
		lcOldOnError = ON("Error")
		ON ERROR 
		MODIFY REPORT (This.aRpts2Run[lnRpt,1])
		ON ERROR &lcOldOnError
	CASE This.oListeners.Count > 0
		TRY
			IF This.cOutput = "PREVIEW"
				REPORT FORM (This.aRpts2Run[lnRpt,1]) OBJECT This.oListeners.Item[1] TO PRINTER PROMPT PREVIEW &tcClauses 
			ELSE 
				REPORT FORM (This.aRpts2Run[lnRpt,1]) OBJECT This.oListeners.Item[1] &tcClauses noconsole
			ENDIF
		CATCH TO oError
			DO ErrLog WITH oError
		ENDTRY 
	OTHERWISE 
		TRY 
			IF This.cOutput = "PREVIEW"
				REPORT FORM (This.aRpts2Run[lnRpt,1]) TO PRINTER PROMPT PREVIEW &tcClauses 
			ELSE 
				REPORT FORM (This.aRpts2Run[lnRpt,1]) &tcClauses noconsole
			ENDIF 
		CATCH TO oError
			DO ErrLog WITH oError
		ENDTRY 
	ENDCASE

	ENDFUNC 
	*-----------------------------------------------------
	FUNCTION RegisterXFRXScripts
	LOCAL lcPages
	* If any scripts were defined in MyXFRXScripts, find them and register them

	This.oScripts = CREATEOBJECT("MyXFRXScripts")

	* Find out what all the native members are so we can pick out the ones that aren't
	AMEMBERS(aNative,This.oScripts,1,"N")
	* Now get all the members
	AMEMBERS(aScripts,This.oScripts,1)
	FOR i = 1 TO ALEN(aScripts,1)
		IF UPPER(aScripts[i,2]) = "METHOD" AND ASCAN(aNative,aScripts[i,1]) = 0
			* It's a method and it's not native
			IF NOT PEMSTATUS(This.oScripts,aScripts[i,1]+"Cond",5) OR EVALUATE(EVALUATE('This.oScripts.'+aScripts[i,1]+"Cond"))
				IF PEMSTATUS(This.oScripts,aScripts[i,1]+"Pages",5)
					* Register it as a page bound script
					lcPages = EVALUATE('This.oScripts.'+aScripts[i,1]+"Pages")
					* The third parameter specifies if the script output should be under
					* on top of report objects.  The fourth parameter specifies the page
					* numbering type--see XFRX manual.
					This.oRpt.RegisterScript(This.oScripts,aScripts[i,1],0,lcPages,0)
				ELSE
					* Register it as a named script
					This.oRpt.RegisterScript(This.oScripts,aScripts[i,1])
				ENDIF 
			ENDIF 
		ENDIF
	ENDFOR 

	ENDFUNC 
	*-----------------------------------------------------
ENDDEFINE
****************************************************************************
* Note the following class has been disabled in this demo, but is an example of using XFRX Scripts if you are interested
* More details are available in the download from http:/www.peisch.com/downloads.html in the file called XFRXSamples.zip
DEFINE CLASS MyXFRXScripts as Custom
	* Create a property for each script that is called the script name plus "Pages".
	* This keeps the RegisterScript function from having to know which pages we want
	* the script to run on.
	* If there is no such property, RegisterXFRXScripts will treat it as a named script.
	* Valid choices are:
	*	"ALL" for all pages
	*	"ODD" for odd pages
	*	"EVEN" for even pages
	*	"X" where X is a specific page number (note that this must be a string, not numeric)
	*	"X,Y,Z" where X, Y and Z are specific pages to print
	* 	"X-Z" where X is the starting page and Z is the ending page
	*	or you can mix the two choices above, e.g. "X,Y,Z,A-D"
	DrawWatermarkPages = "ALL"
	DrawBlueRectanglePages = "5"

	* If a script should only be registered under certain conditions, create a property
	* for the script name plus "Cond" and store the condition in that property.
	* This condition will be evaluated from the RegisterXFRXScripts method of the
	* Output class, so any references to "This" will be referring to that class.
	DrawWatermarkCond = [inlist(this.cOutput,"PREVIEW","PDF")]

	*-----------------------------------------------------
	FUNCTION DrawWatermark
	* Draw a watermark
	LPARAMETERS toXFD
	
	toXFD.SetPos(100,500)
	toXFD.SetColor(210,210,210)
	toXFD.SetFont("Arial",100,.T.)
	* The second parameter of the DrawText method specifies the angle of
	* rotation in a counterclockwise direction.
	* Rotation will only appear for PDF output and when printing
	* from the preview.  Rotation in the preview will be implemented soon.
	toXFD.DrawText("Watermark",45)
	
	ENDFUNC 
	*-----------------------------------------------------
	PROCEDURE drawBlueRectangle
	* This example comes straight out of the XFRX manual
	LPARAMETERS toXFD

	toXFD.setColor(0,0,255,-1,-1,-1)
	toXFD.drawRectangle(100,200,50,50)

	ENDPROC
	*-----------------------------------------------------
ENDDEFINE
*******************************************************
